from dataclasses import dataclass
import imaplib
import email
from email.header import decode_header
import logging
from typing import Iterator
import uuid

logger = logging.getLogger(__name__)


@dataclass
class EmailAttachment:
    filename: str
    mimetype: str
    content: bytes


class Email:
    def __init__(self, id: str, mailbox: str, data: bytes):
        self.id = id
        self.mailbox = mailbox
        self.message = email.message_from_bytes(data)

    @property
    def subject(self) -> str:
        return decode_header(self.message["Subject"])[0][0].decode("utf-8")

    @property
    def body(self) -> str:
        if self.message.is_multipart():
            for part in self.message.walk():
                content_type = part.get_content_type()
                if content_type != "text/plain":
                    continue

                content_disposition = str(part.get("Content-Disposition"))
                if "atttachment" in content_disposition:
                    continue

                charset = part.get_content_charset()
                body_bytes = bytes(part.get_payload(decode=True))
                return body_bytes.decode(charset or "utf-8")

            return ""
        else:
            content_type = self.message.get_content_type()
            if content_type != "text/plain":
                return ""

            body_bytes = bytes(self.message.get_payload(decode=True))
            return body_bytes.decode()

    @property
    def attachments(self) -> Iterator[EmailAttachment]:
        if not self.message.is_multipart():
            return

        for part in self.message.walk():
            content_disposition = part.get("Content-Disposition", None)
            if content_disposition is None or "attachment" not in content_disposition:
                continue

            yield EmailAttachment(
                filename=part.get_filename() or uuid.uuid4().hex,
                mimetype=part.get_content_type(),
                content=bytes(part.get_payload(decode=True)),
            )


class ImapClient:
    def __init__(self, host: str, port: int, username: str, password: str):
        self.host = host
        self.port = port
        self.username = username
        self.password = password

    def __enter__(self):
        self.conn = imaplib.IMAP4_SSL(self.host, self.port)
        self.conn.login(self.username, self.password)
        return self

    def __exit__(self, *args):
        self.conn.close()
        self.conn.logout()

    def get_inbox_mails(self) -> Iterator[Email]:
        MAILBOX = "INBOX"

        self.conn.select(MAILBOX)

        status, data = self.conn.search(None, "ALL")
        if status != "OK":
            raise Exception("Error fetching emails")

        for message_id in data[0].split():
            status, data = self.conn.fetch(message_id, "(RFC822)")
            if status != "OK":
                logger.error(f"Error fetching message {message_id}")
                continue

            if data[0] is None:
                logger.error(f"Message {message_id} is empty")
                continue

            yield Email(
                id=message_id.decode(),
                mailbox=MAILBOX,
                data=bytes(data[0][1]),
            )

    def mark_mail_as_deleted(self, mailbox: str, message_id: str):
        """
        Mark a mail as deleted. The mail will not be deleted until the mailbox
        is expunged.
        """

        self.conn.select(mailbox)
        self.conn.store(message_id, "+FLAGS", "\\Deleted")

    def move_mail(
        self, *, message_id: str, source_mailbox: str, destination_mailbox: str
    ):
        """
        Move a mail from one mailbox to another. The mail will be marked as
        deleted in the source mailbox. Note that the mail will not be deleted
        until the mailbox is expunged.
        """

        self.conn.select(source_mailbox)
        self.conn.copy(message_id, destination_mailbox)
        self.conn.store(message_id, "+FLAGS", "\\Deleted")

    def expunge(self):
        self.conn.expunge()



import logging
import sys
import maildeck.config as config
import maildeck.deck as deck
import maildeck.imap as imap
import maildeck.maildeck as maildeck

logger = logging.getLogger(__name__)


class Command:
    def __init__(self, config: config.Config):
        imap_client = imap.ImapClient(
            host=config.imap_host,
            port=config.imap_port,
            username=config.imap_user,
            password=config.imap_password,
        )

        deck_client = deck.NextcloudDeck(
            base_url=config.nextcloud_base_url,
            username=config.nextcloud_user,
            password=config.nextcloud_password,
        )

        self.maildeck = maildeck.Maildeck(deck_client, imap_client)

        self.nextcloud_board_id = config.nextcloud_board_id
        self.nextcloud_stack_id = config.nextcloud_stack_id

    def run(self):
        self.maildeck.import_emails(
            board_id=self.nextcloud_board_id,
            stack_id=self.nextcloud_stack_id,
        )


def main() -> int:
    try:
        Command(config=config.Config.from_args(sys.argv[1:])).run()
    except Exception as e:
        logger.error(e)
        return 1
    return 0



import argparse
import os
from dataclasses import dataclass, field, fields
from typing import Optional, List, Union, get_args, get_origin


@dataclass
class Config:
    imap_user: str = field(metadata={"help": "IMAP account username."})
    imap_password: str = field(metadata={"help": "IMAP account password."})
    imap_host: str = field(metadata={"help": "IMAP server host address."})

    nextcloud_base_url: str = field(
        metadata={"help": "Base URL for Nextcloud instance."}
    )
    nextcloud_user: str = field(metadata={"help": "Nextcloud account username."})
    nextcloud_password: str = field(metadata={"help": "Nextcloud account password."})
    nextcloud_board_id: int = field(metadata={"help": "Nextcloud board ID."})
    nextcloud_stack_id: Optional[int] = field(
        default=None, metadata={"help": "Nextcloud stack ID. Optional."}
    )

    imap_port: int = field(
        default=993, metadata={"help": "IMAP server port number. Default is 993."}
    )

    @classmethod
    def from_args(cls, argv: List[str]):
        parser = argparse.ArgumentParser(
            description="maildeck - Import emails into Nextcloud Deck.",
            epilog="""
                All arguments can be set as environment variables with the same
                name as the placeholder in this help message.
            """,
        )

        for field_info in fields(cls):
            arg_name = f"--{field_info.name.replace('_', '-')}"
            env_var_name = field_info.name.upper()

            type_origin = get_origin(field_info.type)
            type_args = get_args(field_info.type)
            is_optional = type_origin is Union and type(None) in type_args

            base_type = type_args[0] if is_optional else field_info.type

            default_value = os.getenv(
                env_var_name,
                field_info.default
                if field_info.default != field_info.default_factory
                else None,
            )
            is_required = False if default_value is not None or is_optional else True

            help_message = field_info.metadata.get("help", "")

            parser.add_argument(
                arg_name,
                type=base_type,
                help=help_message,
                default=default_value,
                required=is_required,
            )

        args = parser.parse_args(argv)
        args_dict = {k: v for k, v in vars(args).items() if v is not None}

        return cls(**args_dict)



import json
import urllib.request
import base64
import uuid


class NextcloudDeck:
    API_BASE_PATH = "/index.php/apps/deck/api/v1.1"

    def __init__(self, base_url: str, username: str, password: str):
        self.base_url = base_url

        credentials = f"{username}:{password}"
        encoded_credentials = base64.b64encode(credentials.encode()).decode()
        self.default_headers = {
            "OCS-APIRequest": "true",
            "Authorization": f"Basic {encoded_credentials}",
        }

        self.opener = urllib.request.build_opener()

    def _get(self, path: str) -> dict | list[dict]:
        request = urllib.request.Request(
            self.base_url + self.API_BASE_PATH + path,
            method="GET",
            headers=self.default_headers,
        )

        with self.opener.open(request) as response:
            return json.loads(response.read().decode())

    def _post(self, path: str, data: dict) -> dict:
        headers = self.default_headers.copy()
        headers["Content-Type"] = "application/json"

        request = urllib.request.Request(
            self.base_url + self.API_BASE_PATH + path,
            method="POST",
            headers=headers,
            data=json.dumps(data).encode(),
        )

        with self.opener.open(request) as response:
            return json.loads(response.read().decode())

    def _post_multipart(
        self,
        path: str,
        *,
        data: dict,
        file_field: str,
        file_content: bytes,
        file_filename: str,
        file_mimetype: str,
    ):
        boundary = uuid.uuid4().hex
        headers = self.default_headers.copy()
        headers["Content-Type"] = f"multipart/form-data; boundary={boundary}"

        body = []
        for name, value in data.items():
            body.append(f"--{boundary}".encode("utf-8"))
            body.append(
                f'Content-Disposition: form-data; name="{name}"'.encode("utf-8")
            )
            body.append("".encode("utf-8"))
            body.append(str(value).encode("utf-8"))

        body.append(f"--{boundary}".encode("utf-8"))
        body.append(
            f'Content-Disposition: form-data; name="{file_field}"; filename="{file_filename}"'.encode(
                "utf-8"
            )
        )
        body.append(f"Content-Type: {file_mimetype}".encode("utf-8"))
        body.append("".encode("utf-8"))
        body.append(file_content)

        body.append(f"--{boundary}--".encode("utf-8"))
        body_data = b"\r\n".join(body)

        headers["Content-Length"] = str(len(body_data))

        request = urllib.request.Request(
            self.base_url + self.API_BASE_PATH + path,
            method="POST",
            headers=headers,
            data=body_data,
        )

        with self.opener.open(request) as response:
            return json.loads(response.read().decode())

    def get_stacks(self, board_id: int) -> list[dict]:
        stacks = self._get(f"/boards/{board_id}/stacks")
        if not isinstance(stacks, list):
            raise Exception("Unexpected response")

        return stacks

    def get_first_stack(self, board_id: int) -> dict | None:
        stacks = self.get_stacks(board_id)
        stacks.sort(key=lambda stack: (stack["order"], stack["id"]))

        return stacks[0] if stacks else None

    def create_card(
        self, board_id: int, stack_id: int, title: str, description: str
    ) -> dict:
        data = {
            "title": title,
            "description": description,
            "order": 999,
            "type": "plain",
        }

        return self._post(f"/boards/{board_id}/stacks/{stack_id}/cards", data)

    def create_attachment(
        self,
        *,
        board_id: int,
        stack_id: int,
        card_id: int,
        filename: str,
        mimetype: str,
        content: bytes,
    ) -> dict:
        data = {
            "type": "file",
        }

        return self._post_multipart(
            f"/boards/{board_id}/stacks/{stack_id}/cards/{card_id}/attachments",
            data=data,
            file_field="file",
            file_filename=filename,
            file_mimetype=mimetype,
            file_content=content,
        )



from maildeck.imap import ImapClient
from maildeck.deck import NextcloudDeck


class Maildeck:
    def __init__(self, deck_client: NextcloudDeck, imap_client: ImapClient):
        self.deck_client = deck_client
        self.imap_client = imap_client

    def import_emails(self, board_id: int, stack_id: int | None = None):
        if stack_id is None:
            first_stack = self.deck_client.get_first_stack(board_id)
            stack_id = first_stack["id"] if first_stack else None

        if stack_id is None:
            raise Exception("No stack found")

        with self.imap_client:
            for mail in self.imap_client.get_inbox_mails():
                card = self.deck_client.create_card(
                    board_id=board_id,
                    stack_id=stack_id,
                    title=mail.subject,
                    description=mail.body,
                )

                for attachment in mail.attachments:
                    self.deck_client.create_attachment(
                        board_id=board_id,
                        stack_id=stack_id,
                        card_id=card["id"],
                        filename=attachment.filename,
                        mimetype=attachment.mimetype,
                        content=attachment.content,
                    )

                self.imap_client.mark_mail_as_deleted(mail.mailbox, mail.id)

            self.imap_client.expunge()
